/**********************************************************************
 * SimMips: Simple Computer Simulator of MIPS                         *
 * Copyright (C) 2008-2010 Arch Lab. Tokyo Institute of Technology    *
 *  Dual licensed under New BSD or GPL v2 licenses.                   *
 *  See COPYING file on the base directory for more information.      *
 **********************************************************************/
/* Note: This logger may be accelerated using GCC's -std=c++0x option */

#pragma once
#include <vector>
#include <algorithm>
#ifdef GXX_EXPERIMENTAL_CXX0X
#include <unordered_map>
#else
#include <map>
#endif
#include "stdio.h"

/**********************************************************************/
class BBVector {
 private:
    std::vector<uint032_t> bbaddr;  // index to address
#ifdef GXX_EXPERIMENTAL_CXX0X
    std::unordered_map<uint032_t, int> count; // address to hit count
#else
    std::map<uint032_t, int> count; // address to hit count
#endif
    int delta, total, interval;
    FILE *out;

 public:
    void init(FILE *out = NULL, int interval = 1000000) 
    { this->out = out; this->interval = interval; total = 0; delta = 0; }

    void step(uint032_t addr, int isnop, int isbranch) {
        if (!out) return;
        if (isnop) return;
        delta++;
        if (++total == interval) print();
        if (!isbranch) return;
        if (count.find(addr) == count.end())
            bbaddr.push_back(addr);
        count[addr] += delta;
        delta = 0;
    }

    void print() {
        if (!out) return;
        fprintf(out, "T");
        for (uint i = 0; i < bbaddr.size(); i++) {
            if (count[bbaddr[i]] > 0)
                fprintf(out, ":%d:%d ", i + 1, count[bbaddr[i]]);
            count[bbaddr[i]] = 0;
        }
        fprintf(out, "\n");
        total = 0;
    }

    void debug() {
        if (!out) return;
        for (uint i = 0; i < bbaddr.size(); i++)
            printf("[%d] %08x: %d\n", i + 1, bbaddr[i], count[bbaddr[i]]);
    }
};

/**********************************************************************/














